import logging
import sys

logger = logging.getLogger(__name__)


async def configure_uvloop() -> None:
    """
    Replace the asyncio event loop with uvloop.
    Uvloop is just a faster drop-in replacement. See docs for more info:
    https://uvloop.readthedocs.io/user/index.html
    """
    if sys.platform == 'win32':
        return
    logger.info('Setting uvloop event loop policy')
    import asyncio

    from uvloop import EventLoopPolicy

    asyncio.set_event_loop_policy(EventLoopPolicy())



import logging
from typing import Any

from fastapi import HTTPException, Request
from fastapi.responses import JSONResponse
from fastapi_stack_utils.schemas.http_exceptions import ErrorResponse, ServerError

log = logging.getLogger('fastapi_stack_utils')


def generate_json_response(
    response_body: ErrorResponse, status_code: int, headers: dict[str, str] | None = None
) -> JSONResponse:
    """
    Generate a JSON response with correlation ID attached
    """
    return JSONResponse(content=response_body.dict(), status_code=status_code, headers=headers)


async def http_exception_handler(request: Request, exc: HTTPException) -> JSONResponse:
    """
    Forces the HTTPException output to be of the correct format. Respects status_code raised.
    """
    # We truly have to handle all scenarios, FastAPI allows detail to be `Any`, even though
    # Starlette only allows Optional[str].....
    # https://github.com/tiangolo/fastapi/blob/c5be1b0550f17d827721a5be1dc4344e73b1993f
    # /docs_src/custom_request_and_route/tutorial002.py#L17-L18
    formatted_errors: list[ServerError]
    detail: Any = exc.detail
    if isinstance(detail, (list, tuple)):
        formatted_errors = [ServerError(description=str(element), error=str(element)) for element in detail]
    else:
        formatted_errors = [ServerError(description=str(detail), error=str(detail))]
    return generate_json_response(
        response_body=ErrorResponse(detail=formatted_errors),
        status_code=exc.status_code,
        headers=getattr(exc, 'headers', None) or {},
    )


async def format_and_log_exception_internal(request: Request, exc: Exception) -> JSONResponse:
    """
    Log an error when unhandled exceptions surface an endpoint.
    This exception handler should only be used for internal systems, as it provides the str(exc) in return.
    For customer facing errors, please use `format_and_log_exception_public`
    """
    log.exception('Unhandled exception raised in endpoint: %s', exc)
    response_body = ErrorResponse(detail=[ServerError(description='Internal Server Error', error=str(exc))])
    return generate_json_response(response_body=response_body, status_code=500)


async def format_and_log_exception_public(request: Request, exc: Exception) -> JSONResponse:
    """
    Log an error when unhandled exceptions surface an endpoint.
    This exception handler will hide the actual exception.
    For customer facing errors, please use `format_and_log_exception_public`
    """
    log.exception('Unhandled exception raised in endpoint: %s', exc)
    response_body = ErrorResponse(
        detail=[ServerError(description='Internal Server Error', error='Internal Server Error')]
    )
    return generate_json_response(response_body=response_body, status_code=500)



from datetime import datetime
from logging import Filter, LogRecord
from time import time_ns
from typing import Protocol

import pytz


def get_time_in_nano_seconds() -> str:
    """
    Return timestamp in nanoseconds
    """
    # Get timestamp as nanoseconds
    nano_timestamp = time_ns()
    # Remove nanoseconds and convert into seconds with UTZ format
    date_time = datetime.fromtimestamp(nano_timestamp // 1000000000, tz=pytz.utc)
    return date_time.strftime(f'%Y-%m-%dT%H:%M:%S.{str(int(nano_timestamp % 1000000000)).zfill(9)}')


class NanoStamp(Filter):
    def filter(self, record: LogRecord) -> bool:
        """
        Filter for logging nanoseconds to Logstash on the `nanostamp` field.
        """
        record.nanostamp = get_time_in_nano_seconds()  # type: ignore[attr-defined]
        return True


class Settings(Protocol):
    ENVIRONMENT: str


def generate_base_logging_config(settings: Settings) -> dict:
    """
    Generate a base logging config
    """
    return {
        'version': 1,
        'disable_existing_loggers': False,
        'filters': {
            'correlation_id': {
                '()': 'asgi_correlation_id.CorrelationIdFilter',
                'uuid_length': 8 if settings.ENVIRONMENT == 'dev' else 36,
            },
            'nanostamp': {'()': 'fastapi_stack_utils.logging_config.NanoStamp'},
        },
        'formatters': {
            'console': {
                'format': '%(levelname)-8s  [%(correlation_id)s] %(name)s:%(lineno)d    %(message)s',
            },
            'json': {
                '()': 'pythonjsonlogger.jsonlogger.JsonFormatter',
                'format': """
                    asctime: %(asctime)s
                    created: %(created)f
                    file_name: %(filename)s
                    func_Name: %(funcName)s
                    level_name: %(levelname)s
                    level: %(levelname)s
                    level_no: %(levelno)s
                    lineno: %(lineno)d
                    message: %(message)s
                    module: %(module)s
                    msec: %(msecs)d
                    name: %(name)s
                    pathname: %(pathname)s
                    process: %(process)d
                    process_name: %(processName)s
                    relative_created: %(relativeCreated)d
                    thread: %(thread)d
                    thread_name: %(threadName)s
                    exc_info: %(exc_info)s
                    correlation_id: %(correlation_id)s
                    nanostamp: %(nanostamp)s
                """,
                'datefmt': '%Y-%m-%d %H:%M:%S',
            },
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'filters': ['correlation_id', 'nanostamp'],
                'formatter': 'console',
            },
            'json': {
                'class': 'logging.StreamHandler',
                'filters': ['correlation_id', 'nanostamp'],
                'formatter': 'json',
            },
        },
        'loggers': {
            # third-party packages
            'asgi_correlation_id': {'level': 'WARNING'},
            'fastapi_audit_log': {'level': 'INFO'},
            'gunicorn': {'level': 'INFO'},
            'uvicorn': {'level': 'WARNING'},
        },
        'root': {
            'handlers': ['console'] if settings.ENVIRONMENT in ['dev', 'test'] else ['json'],
            'level': 'DEBUG' if settings.ENVIRONMENT in ['dev', 'test'] else 'INFO',
        },
    }



import contextlib
import logging
from typing import Callable

from fastapi import Request, Response
from fastapi.routing import APIRoute

log = logging.getLogger('fastapi_stack_utils.route')


class AuditLog(APIRoute):
    def get_route_handler(self) -> Callable:
        """
        Overrides `get_route_handler`
        """
        original_route_handler = super().get_route_handler()

        async def input_body_route_handler(request: Request) -> Response:
            """
            Replacement of route_handler that will attempt to log input body
            """
            bytes_body = await request.body()
            extra = {
                'user': request.headers.get('remote-user', 'Unknown'),
                'method': str(request.method),
                'path': str(request.url.path),
                # str(QueryParam) wrongly translates e.g. %20 into `+` instead of `space`
                'query': request.scope['query_string'].decode() if request.query_params else None,
                'str_body': bytes_body.decode() or None,
            }
            with contextlib.suppress(Exception):
                # override body with JSON if possible
                extra['str_body'] = str(await request.json())

            path_param_body = ' | '.join(filter(None, [extra['path'], extra['query'], extra['str_body']]))
            log.info(
                '%s > [%s] | %s',
                extra['user'],
                extra['method'],
                path_param_body,
                extra=extra,
            )
            response: Response = await original_route_handler(request)
            if request.method not in ['OPTIONS', 'GET', 'HEAD']:
                log.info('Response body: %s', response.body.decode())
                log.info('Response headers: %s', response.headers)
            return response

        return input_body_route_handler



import logging
from typing import TYPE_CHECKING, Any

from fastapi import FastAPI

if TYPE_CHECKING:  # pragma: no cover
    from typing import Callable

    from starlette.middleware import Middleware

log = logging.getLogger('fastapi_stack_utils.middleware')


def patch_fastapi_middlewares(middlewares: list['Middleware']) -> None:
    """
    This function will take a list of Middleware(), and replace(extend) the `build_middleware_stack` function in
    FastAPI.
    This means, when you initialize FastAPI(title=...), the `build_new_middleware_stack`-function below
    will be called, instead of `build_middleware_stack` shipped with FastAPI.

    In Starlette applications there are always two middlewares added:
    * `ServerErrorMiddleware` is added as the very outermost middleware, to handle
      any uncaught errors occurring anywhere in the entire stack. (Errors not handled by ExceptionMiddleware)
    * `ExceptionMiddleware` is added as the very innermost middleware, to deal
    with _handled_ exception cases occurring in the routing or endpoints.

    The normal FastAPI middleware stack looks something like this, when using `add_middleware`:
      FastAPI.middleware_stack
        -> ServerErrorMiddleware
          -> CorrelationIdMiddleware
            -> CorsMiddleware
              -> ExceptionMiddleware

    Since the `ServerErrorMiddleware` overrides any uncaught error response, anything we did with the response in
    previous middlewares are overwritten.

    That's why we want this behaviour instead:
      FastAPI.middleware_stack
        -> CorrelationIdMiddleware
          -> CORSMiddleware
            -> ServerErrorMiddleware
              -> ExceptionMiddleware

    Here, we can see that _after_ ServerErrorMiddleware(and ExceptionMiddleware for handled exceptions)
    has done its work, the request will still go through our CorrelationID and CORS middlewares.
    """
    current_middleware_stack = FastAPI.build_middleware_stack

    def _build_new_middleware_stack(self: FastAPI) -> 'Callable[..., Any]':
        """
        Wrap the FastAPI app with given middlewares, while keeping the previously wrapped ones
        """
        app = current_middleware_stack(self)
        for middleware in middlewares:
            app = middleware.cls(app, **middleware.options)
        return app

    FastAPI.build_middleware_stack = _build_new_middleware_stack  # type: ignore



from pydantic import BaseModel, Field


class DefaultError(BaseModel):
    description: str = Field(
        ...,
        title='Informative description of the error',
        description='Equal or more verbose error message',
    )
    error: str = Field(..., title='Object not found', description='Object x does not exist under resource y')


class ServerError(DefaultError):
    ...


class ErrorResponse(BaseModel):
    detail: list[ServerError]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastapi_stack_utils', 'fastapi_stack_utils.schemas']

package_data = \
{'': ['*']}

install_requires = \
['asgi-correlation-id==3.2.0',
 'fastapi==0.85.2',
 'gunicorn==20.1.0',
 'python-json-logger==2.0.4',
 'pytz==2022.6',
 'sentry-sdk==1.10.1',
 'uvicorn[standard]==0.19.0']

setup_kwargs = {
    'name': 'fastapi-stack-utils',
    'version': '0.6.0',
    'description': 'Utils to extend the FastAPI with logging and exception handlers',
    'long_description': 'None',
    'author': 'Jonas Krüger Svensson',
    'author_email': 'jonas-ks@hotmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)
